/* $Id: godi_win32_stubs.c 805 2008-06-11 00:50:03Z gerd $  
 * ----------------------------------------------------------------------
 * Copyright 2008 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
*/

#define WINVER 0x0501
/* Requite at least WinXP SP 1 */


#include <windows.h>
#include <winreg.h>
#include <stdlib.h>
#include <stdio.h>

#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/fail.h>
#include <caml/alloc.h>
#include <caml/callback.h>
#include <caml/custom.h>
#include <caml/signals.h>

/* Stuff from unixsupport.h: */

struct filedescr {
  union {
    HANDLE handle;
    SOCKET socket;
  } fd;
  enum { KIND_HANDLE, KIND_SOCKET } kind;
  int crt_fd;
};

#define Handle_val(v) (((struct filedescr *) Data_custom_val(v))->fd.handle)


CAMLprim value godiwin32_string_of_fd (value h) {
    CAMLparam1(h);
    char buffer[256];
    struct filedescr *fd;

    fd = Data_custom_val(h);
    buffer[0]=0;
    if (fd->kind == KIND_HANDLE) {
	sprintf(buffer, "HANDLE<%ld,%d>", (long) fd->fd.handle, fd->crt_fd);
    }
    else if (fd->kind == KIND_SOCKET) {
	sprintf(buffer, "SOCKET<%ld>", (long) fd->fd.socket);
    }
    else {
	strcat(buffer, "?");
    };

    CAMLreturn(copy_string(buffer));
}



static void registry_error(LONG code) {
    static char buffer[256];

    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
		  FORMAT_MESSAGE_IGNORE_INSERTS,
		  NULL, 
		  code, 
		  0,
		  buffer,
		  sizeof(buffer),
		  NULL);
    caml_raise_with_string(*caml_named_value("Godi_win32.Registry_error"), 
			   buffer);
};


static void registry_str_error(const char *msg) {
    caml_raise_with_string(*caml_named_value("Godi_win32.Registry_error"), 
			   msg);
};


static void sys_error(LONG code, char *path) {
    static char buffer[256];
    int len;

    buffer[0] = 0;
    strncpy(buffer, strerror(code), sizeof(buffer)-1);
    if (path != NULL) {
	len = strlen(buffer);
	strncat(buffer, ": ", sizeof(buffer)-len-1);
	len += 2;
	strncat(buffer, path, sizeof(buffer)-len-1);
    };
    caml_raise_sys_error(caml_copy_string(buffer));
}


struct hkey_struct {
    HKEY hkey;
    int  hkey_closed;
};


#define access_hkey(v) ( (struct hkey_struct *) ( Data_custom_val(v) ) )

static void hkey_finalize(value v) {
    struct hkey_struct *hkeyptr;
    LONG code;

    hkeyptr = access_hkey(v);
    if (!hkeyptr->hkey_closed) {
	code = RegCloseKey(hkeyptr->hkey);  /* but ignore code */
	hkeyptr->hkey_closed = 1;
    }
};


static struct custom_operations hkey_ops = {
    "godiwin32_hkey",
    hkey_finalize,
    custom_compare_default,
    custom_hash_default,
    custom_serialize_default,
    custom_deserialize_default
};


static value create_hkey_value(HKEY hkey) {
    value v;
    struct hkey_struct hks;

    hks.hkey = hkey;
    hks.hkey_closed = 0;
    v = caml_alloc_custom(&hkey_ops, sizeof(hks), 1, 10000);
    memcpy(Data_custom_val(v), &hks, sizeof(hks));
    return v;
};



CAMLprim value godiwin32_get_root_key(value which) {
    HKEY hkey;
    
    switch(Int_val(which)) {
    case 0:
	hkey = HKEY_CLASSES_ROOT; break;
    case 1:
	hkey = HKEY_CURRENT_USER; break;
    case 2:
	hkey = HKEY_LOCAL_MACHINE; break;
    case 3:
	hkey = HKEY_USERS; break;
    default:
	failwith("godiwin32_get_root_key");
    };
    
    return create_hkey_value(hkey);
}

/*
  // RegOpenCurrentUser is missing in Cygwin's version of w32api headers
value godiwin32_reg_open_current_user(value dummy) {
    HKEY hkey;
    LONG code;

    code = RegOpenCurrentUser(KEY_READ, &hkey);
    if (code != ERROR_SUCCESS) registry_error(code);
    return create_hkey_value(hkey);
}
*/

CAMLprim value godiwin32_reg_open_key(value handle, value name) {
    struct hkey_struct *in_hkey;
    HKEY out_hkey;
    LONG code;

    in_hkey = access_hkey(handle);
    if (in_hkey->hkey_closed) registry_str_error("HKEY is closed");
    code = RegOpenKeyEx(in_hkey->hkey, 
			String_val(name), 0, KEY_READ, &out_hkey);
    if (code != ERROR_SUCCESS) registry_error(code);
    return create_hkey_value(out_hkey);
}


CAMLprim value godiwin32_reg_close_key(value handle) {
    struct hkey_struct *hkey;
    LONG code;

    hkey = access_hkey(handle);
    /* Ignore closed keys */
    if (!hkey->hkey_closed) {
	code = RegCloseKey(hkey->hkey);
	if (code != ERROR_SUCCESS) registry_error(code);
	hkey->hkey_closed = 1;
    };
    return Val_unit;
}


CAMLprim value godiwin32_reg_query_value(value handle, value valname) {
    CAMLparam2(handle, valname);
    struct hkey_struct *hkey;
    LONG code;
    DWORD datasize;
    DWORD typetoken;
    long d;
    long long q;
    CAMLlocal2(data, r);

    hkey = access_hkey(handle);
    if (hkey->hkey_closed) registry_str_error("HKEY is closed");
    
    /* First get the size of the data value: */
    code = RegQueryValueEx(hkey->hkey,
			   String_val(valname),
			   0,
			   NULL,
			   NULL,
			   &datasize);
    if (code != ERROR_SUCCESS) registry_error(code);

    /* Allocate the string buffer: */
    data = caml_alloc_string(datasize);
    
    /* Get the data: */
    code = RegQueryValueEx(hkey->hkey,
			   String_val(valname),
			   0,
			   &typetoken,
			   String_val(data),
			   &datasize);
    if (code != ERROR_SUCCESS) registry_error(code);

    /* Construct the result value: */
    switch (typetoken) {
    case REG_BINARY:
	r = caml_alloc(1, 0);
	Field(r, 0) = data;
	break;
    /* case REG_DWORD: */
    case REG_DWORD_LITTLE_ENDIAN:
	d = *( (long *) (String_val(data)));
	r = caml_alloc(1, 1);
	Field(r, 0) = caml_copy_int32(d);
	break;
    case REG_DWORD_BIG_ENDIAN:
	_swab( String_val(data), (char *) &d, 4 );
	r = caml_alloc(1, 1);
	Field(r, 0) = caml_copy_int32(d);
	break;
    case REG_EXPAND_SZ:
	r = caml_alloc(1, 2);
	Field(r, 0) = data;
	break;
    case REG_LINK:
	r = caml_alloc(1, 3);
	Field(r, 0) = data;
	break;
    case REG_MULTI_SZ:
	r = caml_alloc(1, 4);
	Field(r, 0) = data;
	break;
    case REG_NONE:
	r = Val_int(0);
	break;
    /* case REG_QWORD: */
    case REG_QWORD_LITTLE_ENDIAN:
	q = *( (long long *) (String_val(data)));
	r = caml_alloc(1, 5);
	Field(r, 0) = caml_copy_int64(q);
	break;
    case REG_SZ:
	r = caml_alloc(1, 6);
	Field(r, 0) = data;
	break;
    default:
	registry_str_error("Unknown type");
    };

    CAMLreturn(r);
}


CAMLprim value godiwin32_reg_enum_key(value handle) {
    /* Returns the list in reverse order! */
    CAMLparam1(handle);
    struct hkey_struct *hkey;
    int index;
    LONG code;
    char namebuf[256];
    DWORD namelen;
    char classbuf[256];
    DWORD classlen;
    CAMLlocal4(l,ltmp,pair,s);

    hkey = access_hkey(handle);
    if (hkey->hkey_closed) registry_str_error("HKEY is closed");
    
    l = Val_int(0);   /* empty list */
    index = 0;
    code = ERROR_SUCCESS;

    while (code == ERROR_SUCCESS) {
	namelen = sizeof(namebuf);
	classlen = sizeof(classbuf);
	code = RegEnumKeyEx(hkey->hkey,
			    index,
			    namebuf,
			    &namelen,
			    0,
			    classbuf,
			    &classlen,
			    NULL);
	if (code == ERROR_SUCCESS) {
	    pair = caml_alloc(2, 0);
	    s = caml_alloc_string(namelen);
	    memcpy(String_val(s), namebuf, namelen);
	    Store_field(pair, 0, s);
	    s = caml_alloc_string(classlen);
	    memcpy(String_val(s), classbuf, classlen);
	    Store_field(pair, 1, s);
	    ltmp = caml_alloc(2, 0);
	    Store_field(ltmp, 0, pair);
	    Store_field(ltmp, 1, l);
	    l = ltmp;
	    index++;
	};
    };
    
    if (code != ERROR_NO_MORE_ITEMS) registry_error(code);

    CAMLreturn(l);
}


CAMLprim value godiwin32_reg_enum_value(value handle) {
    /* Returns the list in reverse order! */
    CAMLparam1(handle);
    struct hkey_struct *hkey;
    int index;
    LONG code;
    char namebuf[16384];
    DWORD namelen;
    DWORD typetoken;
    CAMLlocal5(l,ltmp,pair,s,t);

    hkey = access_hkey(handle);
    if (hkey->hkey_closed) registry_str_error("HKEY is closed");
    
    l = Val_int(0);   /* empty list */
    index = 0;
    code = ERROR_SUCCESS;

    while (code == ERROR_SUCCESS) {
	namelen = sizeof(namebuf);
	code = RegEnumValue(hkey->hkey,
			    index,
			    namebuf,
			    &namelen,
			    0,
			    &typetoken,
			    NULL,
			    NULL);
	if (code == ERROR_SUCCESS) {
	    pair = caml_alloc(2, 0);
	    s = caml_alloc_string(namelen);
	    memcpy(String_val(s), namebuf, namelen);
	    Store_field(pair, 0, s);

	    switch (typetoken) {
	    case REG_BINARY: t = Val_int(0); break;
	    case REG_DWORD_LITTLE_ENDIAN: t = Val_int(1); break;
	    case REG_DWORD_BIG_ENDIAN: t = Val_int(1); break;
	    case REG_EXPAND_SZ: t = Val_int(2); break;
	    case REG_LINK: t = Val_int(3); break;
	    case REG_MULTI_SZ: t = Val_int(4); break;
	    case REG_NONE: t = Val_int(5); break;
	    case REG_QWORD_LITTLE_ENDIAN: t = Val_int(6); break;
	    case REG_SZ: t = Val_int(7); break;
	    default:
		registry_str_error("Unknown type");
	    };

	    Store_field(pair, 1, t);
	    ltmp = caml_alloc(2, 0);
	    Store_field(ltmp, 0, pair);
	    Store_field(ltmp, 1, l);
	    l = ltmp;
	    index++;
	};
    };
    
    if (code != ERROR_NO_MORE_ITEMS) registry_error(code);

    CAMLreturn(l);
}


CAMLprim value godiwin32_get_logical_drives(value dummy) {
    DWORD drives;
    drives = GetLogicalDrives();
    return Val_long(drives);
}


CAMLprim value godiwin32_get_identity_information(value path) {
    CAMLparam1(path);
    HANDLE h;
    BY_HANDLE_FILE_INFORMATION finfo;
    int code;
    long long index0;
    CAMLlocal3(pair, volume, index);

    h = CreateFile(String_val(path), GENERIC_READ, 
		   FILE_SHARE_READ | FILE_SHARE_WRITE,
		   NULL, OPEN_EXISTING, 
		   FILE_ATTRIBUTE_NORMAL | FILE_FLAG_BACKUP_SEMANTICS, 
		   NULL);
    if (h == INVALID_HANDLE_VALUE) sys_error(GetLastError(), String_val(path));
    code = GetFileInformationByHandle(h, &finfo);
    code = code==0 ? GetLastError() : ERROR_SUCCESS;
    if (!CloseHandle(h))
	sys_error(GetLastError(), NULL);
    if (code != ERROR_SUCCESS)
	sys_error(code, String_val(path));

    volume = caml_copy_int32(finfo.dwVolumeSerialNumber);

    index0 = finfo.nFileIndexLow;
    index0 |= ((long long) finfo.nFileIndexHigh) << 32;
    index = caml_copy_int64(index0);

    pair = caml_alloc(2,0);
    Store_field(pair, 0, volume);
    Store_field(pair, 1, index);

    CAMLreturn(pair);
}


HANDLE sleep_event;
int sleep_init_flag = 0;

static void sleep_init(void) {
    if (!sleep_init_flag) {
	sleep_event = CreateEvent(NULL, 1, 0, NULL);
	sleep_init_flag = 1;
    }
}


CAMLprim value godiwin32_sleep (value d) {
    CAMLparam1(d);
    
    sleep_init();
    ResetEvent(sleep_event);
    caml_enter_blocking_section();
    WaitForSingleObject(sleep_event, Double_val(d) * 1000);
    caml_leave_blocking_section();

    CAMLreturn(Val_unit);
}


CAMLprim value godiwin32_interrupt_sleep(value dummy) {
    sleep_init();
    SetEvent(sleep_event);
    return Val_unit;
}


CAMLprim value godiwin32_generate_ctrl_c(value dummy) {
    GenerateConsoleCtrlEvent(CTRL_C_EVENT, 0);
    return Val_unit;
}


CAMLprim value godiwin32_get_process_id(value phandle) {
    DWORD pid;
    pid = GetProcessId((HANDLE) (Long_val(phandle)));
    if (pid == 0) {
	sys_error(GetLastError(), NULL);
    };
    return Val_long(pid);
}


CAMLprim value godiwin32_peek_pipe(value handle) {
    DWORD total;
    if (!PeekNamedPipe(Handle_val(handle), NULL, 0, NULL, &total, NULL))
	sys_error(GetLastError(), NULL);
    return Val_long(total);
}
